"""This module holds utils to be used by transformations."""

from butterfree.transform.utils.function import Function
from butterfree.transform.utils.window_spec import Window

__all__ = ["Window", "Function"]
